# noinspection PyPep8Naming
import xml.etree.ElementTree as exml

from ..vendor import Qt


# ------------------------------------------------------------------------------
# noinspection PyPep8Naming,PyUnresolvedReferences,PyBroadException
def load(ui_file, base_instance=None):
    try:
        return Qt.QtCompat.loadUi(ui_file, base_instance)

    except:

        # -- For applications such as 3dsmax we have to compile the
        # -- ui differently
        try:
            import pyside2uic as pyuic
            from cStringIO import StringIO

        except:
            raise Exception('No implementation for loadUi found.')

        # -- Read out the xml file
        xml_data = exml.parse(ui_file)

        # -- Get the lcass of the widget and the form
        widget_class = xml_data.find('widget').get('class')
        form_class = xml_data.find('class').text

        # -- Open the ui file as text
        with open(ui_file, 'r') as f:

            # -- Create a file like object
            o = StringIO()
            frame = {}

            # -- Compile the ui into compiled python and execute it
            pyuic.compileUi(f, o, indent=0)
            pyc = compile(o.getvalue(), '<string>', 'exec')
            eval('exec pyc in frame')

            # -- Get the form class
            form_class = frame['Ui_%s' % form_class]

            # -- Alter what we're evaulating based on what version
            # -- of qt we're running
            try:
                base_class = eval('Qt.QtGui.%s' % widget_class)

            except (NameError, AttributeError):
                base_class = eval('Qt.QtWidgets.%s' % widget_class)

        # -- Subclass the loaded classes to build a wrapped widget
        class _WrappedHelper(form_class, base_class):
            # noinspection PyShadowingNames
            def __init__(self, parent=None):
                super(_WrappedHelper, self).__init__(parent)
                self.setupUi(self)

        return _WrappedHelper()



import os
import sys
import json


# -- This is a special environment variable which the user can set in
# -- order to tailor exactly where scribble will write out its data
ENVIRONMENT_VARIABLE = 'PYSCRIBBLE_STORAGE_DIR'

# -- Determine where we should store our data files. This is dependent
# -- upon a couple of factors. Firstly, we allow for hte user to define
# -- an environment variable which specifies the storage location
STORAGE_DIRECTORY = os.environ.get(ENVIRONMENT_VARIABLE, None)

# -- If we have not been provided with a specific storage location
# -- then we need to resolve to a default location, but this is platform
# -- dependent
if sys.platform == 'linux' or sys.platform == 'linux2':
    if 'XDG_CONFIG_HOME' in os.environ:
        STORAGE_DIRECTORY = os.path.join(
            os.environ['XDG_CONFIG_HOME'],
            'pyscribble',
        )

    else:
        STORAGE_DIRECTORY = os.path.join(
            os.environ['HOME'],
            '.config',
            'pyscribble',
        )

elif sys.platform == 'win32':
    STORAGE_DIRECTORY = os.path.join(
        os.environ.get('APPDATA'),
        'pyscribble',
    )

elif sys.platform == 'darwin':
    STORAGE_DIRECTORY = os.path.join(
        os.environ['HOME'],
        'Documents',
        'pyscribble',
    )

else:
    raise Exception(
        (
            '%s is not supported by default. In order to utilise this module '
            'you must define an environment variable (%s) specifying the '
            'storage path'
        ) % (
            sys.platform,
            ENVIRONMENT_VARIABLE,
        )
    )


# ------------------------------------------------------------------------------
def get(identifier, *args, **kwargs):
    """
    This will attempt to retrieve a Scribble Dictionary from the 
    STORAGE_DIRECTORY. If it does not exist, an empty Scribble 
    Dictionary will be created which will be savable with the given
    identifier.
    
    :param identifier: Unique identifier to access a specific scribble
        dictionary.
    :type identifier: str
    
    :return: ScribbleDictionary
    """
    return ScribbleDictionary(identifier, *args, **kwargs)


# ------------------------------------------------------------------------------
class ScribbleDictionary(dict):

    # --------------------------------------------------------------------------
    def __init__(self, identifier, *args, **kwargs):
        super(ScribbleDictionary, self).__init__(*args, **kwargs)

        # -- Store the identifier, as this is used whenever the
        # -- scribble dictionary is stored.
        self.identifier = identifier

        # -- Initiate a load of any persistent data for the given
        # -- identifier
        self.load()

    # --------------------------------------------------------------------------
    def load(self):
        """
        Loads the data from the disk if it exists and updates this
        dictionary.
        """
        if os.path.exists(self.location()):
            with open(self.location(), 'r') as f:
                self.update(json.load(f))

    # --------------------------------------------------------------------------
    def save(self):
        """
        Saves the ScribbleDictionary data to a persistent state
        """
        # -- Ensure the location to save to exists, otherwise the
        # -- write will fail
        if not os.path.exists(STORAGE_DIRECTORY):
            os.makedirs(STORAGE_DIRECTORY)

        # -- Serialise to json - we wrap this in a try/except in case
        # -- the data is not json serialisable.
        try:
            data = json.dumps(
                self,
                indent=4,
                sort_keys=True,
            )

        except BaseException:
            raise Exception(
                'Could not encode the data within the Scribble Dictionary '
                'to JSON. Please ensure any stored data can be serialised '
                'to JSON.'
            )

        # -- Write the data out
        with open(self.location(), 'w') as f:
            f.write(data)

    # --------------------------------------------------------------------------
    def location(self):
        """
        Returns the location of the persistent file. This is regardless of
        whether the file exists or not.
        
        :return: str
        """
        return os.path.join(
            STORAGE_DIRECTORY,
            '%s.json' % (
                self.identifier,
            ),
        )



from ..vendor import Qt


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences
def printEventName(event):
    """
    Prints the name of the event type being given.

    :param event: The event to print the type of
    :type event: QEvent

    :return: None
    """
    if hasattr(Qt.QEvent, 'AccessibilityDescription'):
        if event.type() == Qt.QEvent.AccessibilityDescription:
            print('AccessibilityDescription')
    if hasattr(Qt.QEvent, 'AccessibilityHelp'):
        if event.type() == Qt.QEvent.AccessibilityHelp:
            print('AccessibilityHelp')
    if hasattr(Qt.QEvent, 'AccessibilityPrepare'):
        if event.type() == Qt.QEvent.AccessibilityPrepare:
            print('AccessibilityPrepare')
    if hasattr(Qt.QEvent, 'ActionAdded'):
        if event.type() == Qt.QEvent.ActionAdded:
            print('ActionAdded')
    if hasattr(Qt.QEvent, 'ActionChanged'):
        if event.type() == Qt.QEvent.ActionChanged:
            print('ActionChanged')
    if hasattr(Qt.QEvent, 'ActionRemoved'):
        if event.type() == Qt.QEvent.ActionRemoved:
            print('ActionRemoved')
    if hasattr(Qt.QEvent, 'ActivationChange'):
        if event.type() == Qt.QEvent.ActivationChange:
            print('ActivationChange')
    if hasattr(Qt.QEvent, 'ApplicationActivate'):
        if event.type() == Qt.QEvent.ApplicationActivate:
            print('ApplicationActivate')
    if hasattr(Qt.QEvent, 'ApplicationActivated'):
        if event.type() == Qt.QEvent.ApplicationActivated:
            print('ApplicationActivated')
    if hasattr(Qt.QEvent, 'ApplicationDeactivate'):
        if event.type() == Qt.QEvent.ApplicationDeactivate:
            print('ApplicationDeactivate')
    if hasattr(Qt.QEvent, 'ApplicationFontChange'):
        if event.type() == Qt.QEvent.ApplicationFontChange:
            print('ApplicationFontChange')
    if hasattr(Qt.QEvent, 'ApplicationLayoutDirectionChange'):
        if event.type() == Qt.QEvent.ApplicationLayoutDirectionChange:
            print('ApplicationLayoutDirectionChange')
    if hasattr(Qt.QEvent, 'ApplicationPaletteChange'):
        if event.type() == Qt.QEvent.ApplicationPaletteChange:
            print('ApplicationPaletteChange')
    if hasattr(Qt.QEvent, 'ApplicationWindowIconChange'):
        if event.type() == Qt.QEvent.ApplicationWindowIconChange:
            print('ApplicationWindowIconChange')
    if hasattr(Qt.QEvent, 'ChildAdded'):
        if event.type() == Qt.QEvent.ChildAdded:
            print('ChildAdded')
    if hasattr(Qt.QEvent, 'ChildInserted'):
        if event.type() == Qt.QEvent.ChildInserted:
            print('ChildInserted')
    if hasattr(Qt.QEvent, 'ChildPolished'):
        if event.type() == Qt.QEvent.ChildPolished:
            print('ChildPolished')
    if hasattr(Qt.QEvent, 'ChildRemoved'):
        if event.type() == Qt.QEvent.ChildRemoved:
            print('ChildRemoved')
    if hasattr(Qt.QEvent, 'Clipboard'):
        if event.type() == Qt.QEvent.Clipboard:
            print('Clipboard')
    if hasattr(Qt.QEvent, 'Close'):
        if event.type() == Qt.QEvent.Close:
            print('Close')
    if hasattr(Qt.QEvent, 'CloseSoftwareInputPanel'):
        if event.type() == Qt.QEvent.CloseSoftwareInputPanel:
            print('CloseSoftwareInputPanel')
    if hasattr(Qt.QEvent, 'ContentsRectChange'):
        if event.type() == Qt.QEvent.ContentsRectChange:
            print('ContentsRectChange')
    if hasattr(Qt.QEvent, 'ContextMenu'):
        if event.type() == Qt.QEvent.ContextMenu:
            print('ContextMenu')
    if hasattr(Qt.QEvent, 'CursorChange'):
        if event.type() == Qt.QEvent.CursorChange:
            print('CursorChange')
    if hasattr(Qt.QEvent, 'DeferredDelete'):
        if event.type() == Qt.QEvent.DeferredDelete:
            print('DeferredDelete')
    if hasattr(Qt.QEvent, 'DragEnter'):
        if event.type() == Qt.QEvent.DragEnter:
            print('DragEnter')
    if hasattr(Qt.QEvent, 'DragLeave'):
        if event.type() == Qt.QEvent.DragLeave:
            print('DragLeave')
    if hasattr(Qt.QEvent, 'DragMove'):
        if event.type() == Qt.QEvent.DragMove:
            print('DragMove')
    if hasattr(Qt.QEvent, 'Drop'):
        if event.type() == Qt.QEvent.Drop:
            print('Drop')
    if hasattr(Qt.QEvent, 'EnabledChange'):
        if event.type() == Qt.QEvent.EnabledChange:
            print('EnabledChange')
    if hasattr(Qt.QEvent, 'Enter'):
        if event.type() == Qt.QEvent.Enter:
            print('Enter')
    if hasattr(Qt.QEvent, 'EnterEditFocus'):
        if event.type() == Qt.QEvent.EnterEditFocus:
            print('EnterEditFocus')
    if hasattr(Qt.QEvent, 'EnterWhatsThisMode'):
        if event.type() == Qt.QEvent.EnterWhatsThisMode:
            print('EnterWhatsThisMode')
    if hasattr(Qt.QEvent, 'FileOpen'):
        if event.type() == Qt.QEvent.FileOpen:
            print('FileOpen')
    if hasattr(Qt.QEvent, 'FocusIn'):
        if event.type() == Qt.QEvent.FocusIn:
            print('FocusIn')
    if hasattr(Qt.QEvent, 'FocusOut'):
        if event.type() == Qt.QEvent.FocusOut:
            print('FocusOut')
    if hasattr(Qt.QEvent, 'FontChange'):
        if event.type() == Qt.QEvent.FontChange:
            print('FontChange')
    if hasattr(Qt.QEvent, 'GrabKeyboard'):
        if event.type() == Qt.QEvent.GrabKeyboard:
            print('GrabKeyboard')
    if hasattr(Qt.QEvent, 'GrabMouse'):
        if event.type() == Qt.QEvent.GrabMouse:
            print('GrabMouse')
    if hasattr(Qt.QEvent, 'GraphicsSceneContextMenu'):
        if event.type() == Qt.QEvent.GraphicsSceneContextMenu:
            print('GraphicsSceneContextMenu')
    if hasattr(Qt.QEvent, 'GraphicsSceneDragEnter'):
        if event.type() == Qt.QEvent.GraphicsSceneDragEnter:
            print('GraphicsSceneDragEnter')
    if hasattr(Qt.QEvent, 'GraphicsSceneDragLeave'):
        if event.type() == Qt.QEvent.GraphicsSceneDragLeave:
            print('GraphicsSceneDragLeave')
    if hasattr(Qt.QEvent, 'GraphicsSceneDragMove'):
        if event.type() == Qt.QEvent.GraphicsSceneDragMove:
            print('GraphicsSceneDragMove')
    if hasattr(Qt.QEvent, 'GraphicsSceneDrop'):
        if event.type() == Qt.QEvent.GraphicsSceneDrop:
            print('GraphicsSceneDrop')
    if hasattr(Qt.QEvent, 'GraphicsSceneHelp'):
        if event.type() == Qt.QEvent.GraphicsSceneHelp:
            print('GraphicsSceneHelp')
    if hasattr(Qt.QEvent, 'GraphicsSceneHoverEnter'):
        if event.type() == Qt.QEvent.GraphicsSceneHoverEnter:
            print('GraphicsSceneHoverEnter')
    if hasattr(Qt.QEvent, 'GraphicsSceneHoverLeave'):
        if event.type() == Qt.QEvent.GraphicsSceneHoverLeave:
            print('GraphicsSceneHoverLeave')
    if hasattr(Qt.QEvent, 'GraphicsSceneHoverMove'):
        if event.type() == Qt.QEvent.GraphicsSceneHoverMove:
            print('GraphicsSceneHoverMove')
    if hasattr(Qt.QEvent, 'GraphicsSceneMouseDoubleClick'):
        if event.type() == Qt.QEvent.GraphicsSceneMouseDoubleClick:
            print('GraphicsSceneMouseDoubleClick')
    if hasattr(Qt.QEvent, 'GraphicsSceneMouseMove'):
        if event.type() == Qt.QEvent.GraphicsSceneMouseMove:
            print('GraphicsSceneMouseMove')
    if hasattr(Qt.QEvent, 'GraphicsSceneMousePress'):
        if event.type() == Qt.QEvent.GraphicsSceneMousePress:
            print('GraphicsSceneMousePress')
    if hasattr(Qt.QEvent, 'GraphicsSceneMouseRelease'):
        if event.type() == Qt.QEvent.GraphicsSceneMouseRelease:
            print('GraphicsSceneMouseRelease')
    if hasattr(Qt.QEvent, 'GraphicsSceneMove'):
        if event.type() == Qt.QEvent.GraphicsSceneMove:
            print('GraphicsSceneMove')
    if hasattr(Qt.QEvent, 'GraphicsSceneResize'):
        if event.type() == Qt.QEvent.GraphicsSceneResize:
            print('GraphicsSceneResize')
    if hasattr(Qt.QEvent, 'GraphicsSceneWheel'):
        if event.type() == Qt.QEvent.GraphicsSceneWheel:
            print('GraphicsSceneWheel')
    if hasattr(Qt.QEvent, 'Hide'):
        if event.type() == Qt.QEvent.Hide:
            print('Hide')
    if hasattr(Qt.QEvent, 'HideToParent'):
        if event.type() == Qt.QEvent.HideToParent:
            print('HideToParent')
    if hasattr(Qt.QEvent, 'HoverEnter'):
        if event.type() == Qt.QEvent.HoverEnter:
            print('HoverEnter')
    if hasattr(Qt.QEvent, 'HoverLeave'):
        if event.type() == Qt.QEvent.HoverLeave:
            print('HoverLeave')
    if hasattr(Qt.QEvent, 'HoverMove'):
        if event.type() == Qt.QEvent.HoverMove:
            print('HoverMove')
    if hasattr(Qt.QEvent, 'IconDrag'):
        if event.type() == Qt.QEvent.IconDrag:
            print('IconDrag')
    if hasattr(Qt.QEvent, 'IconTextChange'):
        if event.type() == Qt.QEvent.IconTextChange:
            print('IconTextChange')
    if hasattr(Qt.QEvent, 'InputMethod'):
        if event.type() == Qt.QEvent.InputMethod:
            print('InputMethod')
    if hasattr(Qt.QEvent, 'KeyPress'):
        if event.type() == Qt.QEvent.KeyPress:
            print('KeyPress')
    if hasattr(Qt.QEvent, 'KeyRelease'):
        if event.type() == Qt.QEvent.KeyRelease:
            print('KeyRelease')
    if hasattr(Qt.QEvent, 'LanguageChange'):
        if event.type() == Qt.QEvent.LanguageChange:
            print('LanguageChange')
    if hasattr(Qt.QEvent, 'LayoutDirectionChange'):
        if event.type() == Qt.QEvent.LayoutDirectionChange:
            print('LayoutDirectionChange')
    if hasattr(Qt.QEvent, 'LayoutRequest'):
        if event.type() == Qt.QEvent.LayoutRequest:
            print('LayoutRequest')
    if hasattr(Qt.QEvent, 'Leave'):
        if event.type() == Qt.QEvent.Leave:
            print('Leave')
    if hasattr(Qt.QEvent, 'LeaveEditFocus'):
        if event.type() == Qt.QEvent.LeaveEditFocus:
            print('LeaveEditFocus')
    if hasattr(Qt.QEvent, 'LeaveWhatsThisMode'):
        if event.type() == Qt.QEvent.LeaveWhatsThisMode:
            print('LeaveWhatsThisMode')
    if hasattr(Qt.QEvent, 'LocaleChange'):
        if event.type() == Qt.QEvent.LocaleChange:
            print('LocaleChange')
    if hasattr(Qt.QEvent, 'NonClientAreaMouseButtonDblClick'):
        if event.type() == Qt.QEvent.NonClientAreaMouseButtonDblClick:
            print('NonClientAreaMouseButtonDblClick')
    if hasattr(Qt.QEvent, 'NonClientAreaMouseButtonPress'):
        if event.type() == Qt.QEvent.NonClientAreaMouseButtonPress:
            print('NonClientAreaMouseButtonPress')
    if hasattr(Qt.QEvent, 'NonClientAreaMouseButtonRelease'):
        if event.type() == Qt.QEvent.NonClientAreaMouseButtonRelease:
            print('NonClientAreaMouseButtonRelease')
    if hasattr(Qt.QEvent, 'NonClientAreaMouseMove'):
        if event.type() == Qt.QEvent.NonClientAreaMouseMove:
            print('NonClientAreaMouseMove')
    if hasattr(Qt.QEvent, 'MacSizeChange'):
        if event.type() == Qt.QEvent.MacSizeChange:
            print('MacSizeChange')
    if hasattr(Qt.QEvent, 'MenubarUpdated'):
        if event.type() == Qt.QEvent.MenubarUpdated:
            print('MenubarUpdated')
    if hasattr(Qt.QEvent, 'MetaCall'):
        if event.type() == Qt.QEvent.MetaCall:
            print('MetaCall')
    if hasattr(Qt.QEvent, 'ModifiedChange'):
        if event.type() == Qt.QEvent.ModifiedChange:
            print('ModifiedChange')
    if hasattr(Qt.QEvent, 'MouseButtonDblClick'):
        if event.type() == Qt.QEvent.MouseButtonDblClick:
            print('MouseButtonDblClick')
    if hasattr(Qt.QEvent, 'MouseButtonPress'):
        if event.type() == Qt.QEvent.MouseButtonPress:
            print('MouseButtonPress')
    if hasattr(Qt.QEvent, 'MouseButtonRelease'):
        if event.type() == Qt.QEvent.MouseButtonRelease:
            print('MouseButtonRelease')
    if hasattr(Qt.QEvent, 'MouseMove'):
        if event.type() == Qt.QEvent.MouseMove:
            print('MouseMove')
    if hasattr(Qt.QEvent, 'MouseTrackingChange'):
        if event.type() == Qt.QEvent.MouseTrackingChange:
            print('MouseTrackingChange')
    if hasattr(Qt.QEvent, 'Move'):
        if event.type() == Qt.QEvent.Move:
            print('Move')
    if hasattr(Qt.QEvent, 'Paint'):
        if event.type() == Qt.QEvent.Paint:
            print('Paint')
    if hasattr(Qt.QEvent, 'PaletteChange'):
        if event.type() == Qt.QEvent.PaletteChange:
            print('PaletteChange')
    if hasattr(Qt.QEvent, 'ParentAboutToChange'):
        if event.type() == Qt.QEvent.ParentAboutToChange:
            print('ParentAboutToChange')
    if hasattr(Qt.QEvent, 'ParentChange'):
        if event.type() == Qt.QEvent.ParentChange:
            print('ParentChange')
    if hasattr(Qt.QEvent, 'Polish'):
        if event.type() == Qt.QEvent.Polish:
            print('Polish')
    if hasattr(Qt.QEvent, 'PolishRequest'):
        if event.type() == Qt.QEvent.PolishRequest:
            print('PolishRequest')
    if hasattr(Qt.QEvent, 'QueryWhatsThis'):
        if event.type() == Qt.QEvent.QueryWhatsThis:
            print('QueryWhatsThis')
    if hasattr(Qt.QEvent, 'RequestSoftwareInputPanel'):
        if event.type() == Qt.QEvent.RequestSoftwareInputPanel:
            print('RequestSoftwareInputPanel')
    if hasattr(Qt.QEvent, 'Resize'):
        if event.type() == Qt.QEvent.Resize:
            print('Resize')
    if hasattr(Qt.QEvent, 'Shortcut'):
        if event.type() == Qt.QEvent.Shortcut:
            print('Shortcut')
    if hasattr(Qt.QEvent, 'ShortcutOverride'):
        if event.type() == Qt.QEvent.ShortcutOverride:
            print('ShortcutOverride')
    if hasattr(Qt.QEvent, 'Show'):
        if event.type() == Qt.QEvent.Show:
            print('Show')
    if hasattr(Qt.QEvent, 'ShowToParent'):
        if event.type() == Qt.QEvent.ShowToParent:
            print('ShowToParent')
    if hasattr(Qt.QEvent, 'SockAct'):
        if event.type() == Qt.QEvent.SockAct:
            print('SockAct')
    if hasattr(Qt.QEvent, 'StateMachineSignal'):
        if event.type() == Qt.QEvent.StateMachineSignal:
            print('StateMachineSignal')
    if hasattr(Qt.QEvent, 'StateMachineWrapped'):
        if event.type() == Qt.QEvent.StateMachineWrapped:
            print('StateMachineWrapped')
    if hasattr(Qt.QEvent, 'StatusTip'):
        if event.type() == Qt.QEvent.StatusTip:
            print('StatusTip')
    if hasattr(Qt.QEvent, 'StyleChange'):
        if event.type() == Qt.QEvent.StyleChange:
            print('StyleChange')
    if hasattr(Qt.QEvent, 'TabletMove'):
        if event.type() == Qt.QEvent.TabletMove:
            print('TabletMove')
    if hasattr(Qt.QEvent, 'TabletPress'):
        if event.type() == Qt.QEvent.TabletPress:
            print('TabletPress')
    if hasattr(Qt.QEvent, 'TabletRelease'):
        if event.type() == Qt.QEvent.TabletRelease:
            print('TabletRelease')
    if hasattr(Qt.QEvent, 'OkRequest'):
        if event.type() == Qt.QEvent.OkRequest:
            print('OkRequest')
    if hasattr(Qt.QEvent, 'TabletEnterProximity'):
        if event.type() == Qt.QEvent.TabletEnterProximity:
            print('TabletEnterProximity')
    if hasattr(Qt.QEvent, 'TabletLeaveProximity'):
        if event.type() == Qt.QEvent.TabletLeaveProximity:
            print('TabletLeaveProximity')
    if hasattr(Qt.QEvent, 'Timer'):
        if event.type() == Qt.QEvent.Timer:
            print('Timer')
    if hasattr(Qt.QEvent, 'ToolBarChange'):
        if event.type() == Qt.QEvent.ToolBarChange:
            print('ToolBarChange')
    if hasattr(Qt.QEvent, 'ToolTip'):
        if event.type() == Qt.QEvent.ToolTip:
            print('ToolTip')
    if hasattr(Qt.QEvent, 'ToolTipChange'):
        if event.type() == Qt.QEvent.ToolTipChange:
            print('ToolTipChange')
    if hasattr(Qt.QEvent, 'UngrabKeyboard'):
        if event.type() == Qt.QEvent.UngrabKeyboard:
            print('UngrabKeyboard')
    if hasattr(Qt.QEvent, 'UngrabMouse'):
        if event.type() == Qt.QEvent.UngrabMouse:
            print('UngrabMouse')
    if hasattr(Qt.QEvent, 'UpdateLater'):
        if event.type() == Qt.QEvent.UpdateLater:
            print('UpdateLater')
    if hasattr(Qt.QEvent, 'UpdateRequest'):
        if event.type() == Qt.QEvent.UpdateRequest:
            print('UpdateRequest')
    if hasattr(Qt.QEvent, 'WhatsThis'):
        if event.type() == Qt.QEvent.WhatsThis:
            print('WhatsThis')
    if hasattr(Qt.QEvent, 'WhatsThisClicked'):
        if event.type() == Qt.QEvent.WhatsThisClicked:
            print('WhatsThisClicked')
    if hasattr(Qt.QEvent, 'Wheel'):
        if event.type() == Qt.QEvent.Wheel:
            print('Wheel')
    if hasattr(Qt.QEvent, 'WinEventAct'):
        if event.type() == Qt.QEvent.WinEventAct:
            print('WinEventAct')
    if hasattr(Qt.QEvent, 'WindowActivate'):
        if event.type() == Qt.QEvent.WindowActivate:
            print('WindowActivate')
    if hasattr(Qt.QEvent, 'WindowBlocked'):
        if event.type() == Qt.QEvent.WindowBlocked:
            print('WindowBlocked')
    if hasattr(Qt.QEvent, 'WindowDeactivate'):
        if event.type() == Qt.QEvent.WindowDeactivate:
            print('WindowDeactivate')
    if hasattr(Qt.QEvent, 'WindowIconChange'):
        if event.type() == Qt.QEvent.WindowIconChange:
            print('WindowIconChange')
    if hasattr(Qt.QEvent, 'WindowStateChange'):
        if event.type() == Qt.QEvent.WindowStateChange:
            print('WindowStateChange')
    if hasattr(Qt.QEvent, 'WindowTitleChange'):
        if event.type() == Qt.QEvent.WindowTitleChange:
            print('WindowTitleChange')
    if hasattr(Qt.QEvent, 'WindowUnblocked'):
        if event.type() == Qt.QEvent.WindowUnblocked:
            print('WindowUnblocked')
    if hasattr(Qt.QEvent, 'ZOrderChange'):
        if event.type() == Qt.QEvent.ZOrderChange:
            print('ZOrderChange')
    if hasattr(Qt.QEvent, 'KeyboardLayoutChange'):
        if event.type() == Qt.QEvent.KeyboardLayoutChange:
            print('KeyboardLayoutChange')
    if hasattr(Qt.QEvent, 'DynamicPropertyChange'):
        if event.type() == Qt.QEvent.DynamicPropertyChange:
            print('DynamicPropertyChange')
    if hasattr(Qt.QEvent, 'TouchBegin'):
        if event.type() == Qt.QEvent.TouchBegin:
            print('TouchBegin')
    if hasattr(Qt.QEvent, 'TouchUpdate'):
        if event.type() == Qt.QEvent.TouchUpdate:
            print('TouchUpdate')
    if hasattr(Qt.QEvent, 'TouchEnd'):
        if event.type() == Qt.QEvent.TouchEnd:
            print('TouchEnd')
    if hasattr(Qt.QEvent, 'WinIdChange'):
        if event.type() == Qt.QEvent.WinIdChange:
            print('WinIdChange')
    if hasattr(Qt.QEvent, 'Gesture'):
        if event.type() == Qt.QEvent.Gesture:
            print('Gesture')
    if hasattr(Qt.QEvent, 'GestureOverride'):
        if event.type() == Qt.QEvent.GestureOverride:
            print('GestureOverride')



"""
Scribble is a mechanism to store dictionary information to a persistent
state. This allows the data to be easily retrieved and edited, making it
a good option when trying to store settings data. 

Note: Data saved within the ScribbleDictionary must be json serialisable.

An example of creating a persisten data block would be:

..code-block:: python

    import os
    import scribble
    
    
    # -- Instance a scribble dictionary. We can treat this exactly
    # -- as we treat a dictionary
    data = scribble.get('foobar')
    
    # -- Set some key/value pairs in the ScribbleDictionary
    data['option_a'] = True
    data['option_b'] = 123
    
    # -- Calling .save() will trigger the dictionary
    # -- to store its current state to a persistent state
    data.save()
    
    # -- We can see the location the data is saved to, and we can
    # -- see that it does indeed exist
    print(data.location())
    print(os.path.exists(data.location()))
    
    
Equally, we can re-retrieve that data in a completely new instance of python
using the following code:

..code-block:: python

    # -- Instance a scribble dictionary with the same identifier
    data = scribble.get('foobar')
    
    # -- Print the fact that we have retrieved the same 
    # -- infromation
    print(data)
    # {'option_a': True, 'option_b': 123}
    
    # -- We can then further edit the data
    data['option_a'] = False
    
    # -- Calling .save() will trigger the dictionary
    # -- to store its current state to a persistent state
    data.save()
    
By default scribble will store its data in the follow platform specific
locations:

    * Windows: %APPDATA%/pyscribble
    
    * Linux: %XDG_CONFIG_HOME%/pyscribble if the environment variable exists, otherwise %HOME%/.config/pyscribble
    
    * OSX: Not yet supported

However, you can override these paths by setting an environment variable
PYSCRIBBLE_STORAGE_DIR, if this is set then the path defined by that variable
will always be used over the default behaviour.
"""
from .core import get



"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
FlowLayout
Custom layout that mimics the behaviour of a flow layout (not supported in PyQt by default)
Just added comments on the offical PySide example.
@date 08-2013
@source http://josbalcaen.com/pyqt-flowlayout-maya-python/
"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
from ..vendor import Qt



class FlowLayout(Qt.QtWidgets.QLayout):
    """Custom layout that mimics the behaviour of a flow layout"""

    def __init__(self, parent=None, margin=0, spacing=-1):
        super(FlowLayout, self).__init__(parent)

        self._margin = 0
        # Set margin and spacing
        if parent is not None:
            self.setMargin(margin)

        self.setSpacing(spacing)

        self.itemList = []

    def margin(self):
        return self._margin

    def setMargin(self, value):
        self._margin = value

    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addItem(self, item):
        self.itemList.append(item)

    def count(self):
        return len(self.itemList)

    def itemAt(self, index):
        if index >= 0 and index < len(self.itemList):
            return self.itemList[index]
        return None

    def takeAt(self, index):
        if index >= 0 and index < len(self.itemList):
            return self.itemList.pop(index)
        return None

    def insertWidget(self, index, widget):
        item = Qt.QtWidgets.QWidgetItem(widget)
        self.itemList.insert(index, item)

    def expandingDirections(self):
        return Qt.QtCore.Qt.Orientations(Qt.QtCore.Qt.Horizontal)

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        height = self.doLayout(Qt.QtCore.QRect(0, 0, width, 0), True)
        return height

    def setGeometry(self, rect):
        super(FlowLayout, self).setGeometry(rect)
        self.doLayout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        # Calculate the size
        size = Qt.QtCore.QSize()

        for item in self.itemList:
            size = size.expandedTo(item.minimumSize())

        # Add the margins
        size += Qt.QtCore.QSize(2 * self.margin(), 2 * self.margin())

        return size

    def doLayout(self, rect, testOnly):
        x = rect.x()
        y = rect.y()
        lineHeight = 0

        for item in self.itemList:
            wid = item.widget()
            spaceX = self.spacing()
            spaceY = self.spacing()
            nextX = x + item.sizeHint().width() + spaceX
            if nextX - spaceX > rect.right() and lineHeight > 0:
                x = rect.x()
                y = y + lineHeight + spaceY
                nextX = x + item.sizeHint().width() + spaceX
                lineHeight = 0

            if not testOnly:
                item.setGeometry(Qt.QtCore.QRect(Qt.QtCore.QPoint(x, y), item.sizeHint()))

            x = nextX
            lineHeight = max(lineHeight, item.sizeHint().height())

        return y + lineHeight - rect.y()



from .. import utilities
from .. import extensions
from .. import applyStyle


# ------------------------------------------------------------------------------
def quick_app(window_title, style=None):
    """
    Decorator to use on a function that is expected to return a widget that
    will be set as the Main Widget in the window that is created.

    :param window_title: App and window title.
    :type window_title: str
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            q_app = utilities.qApp()

            widget = func(*args, **kwargs)

            window = extensions.windows.MemorableWindow(
                identifier=window_title,
                parent=utilities.windows.mainWindow(),
            )

            window.setCentralWidget(widget)

            # -- Set the window properties
            window.setWindowTitle(window_title)

            if style:
                applyStyle(
                    style,
                    apply_to=window
                )

            # -- Show the ui, and if we're blocking call the exec_
            window.show()

            q_app.exec_()

            return window

        return wrapper

    return decorator



"""
This module holds functionality to apply and generate styling information
to widgets
"""
import os
import re

from . import toList
from .. import constants


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences
def apply(styles, apply_to, **kwargs):
    """
    Applies all the given styles in order (meaning each will override any
    overlapping elements of the previous).

    All kwargs are considered replacements when wrangling your stylesheet.
    This allows you to place variables into your css file and have them resolved
    at runtime.

    :param styles: This can be a single stylesheet or a list of stylesheets.
        The stylesheet can be given in three different forms (and is checked
        in this order):

            * An absolute filepath to a css/qss file

            * A Name of any stylesheet which exists in any location defined
                within the QUTE_STYLE_PATH environment variable

            * Actual stylesheet data
    :type styles: This can be given either a single string or a list of
        strings

    :param apply_to: The QWidget which should have the stylesheet applied
    :type apply_to: QWidget

    :return: None


    .. code-block:: python

        >>> import qute as sk
        >>>
        >>> # -- Ensure we have an application
        >>> if not sk.QtWidgets.QApplication.instance():
        ...     q_app = sk.QtWidgets.QApplication([])
        >>>
        >>> # -- Create a widget
        >>> button = sk.QtWidgets.QPushButton()
        >>> sk.styles.applyStyle("flame", button)

    """
    # -- As we allow single elements or a list of elements to be
    # -- passed we need to conform that now
    styles = toList(styles)
    available_styles = None

    # -- Start collating our style data
    compounded_style = ''

    for given_style in styles:

        extracted_style = None

        # -- Firstly we check if we're given an absolute path which
        # -- resolves
        if os.path.exists(given_style):
            with open(given_style, 'r') as f:
                extracted_style = f.read()

        # -- If we did not extract a style we need to check the locations
        # -- in our environment path
        if not extracted_style:

            # -- Read out the available styles if it is our first iteration
            # -- during this call
            available_styles = available_styles or _getAvailableStyles()

            # -- If we have the given style we store it
            if given_style in available_styles:
                with open(available_styles[given_style], 'r') as f:
                    extracted_style = f.read()

        # -- Finally we need to check if we think this might be a
        # -- style in its own right
        if not extracted_style and ';' in given_style:
            extracted_style = given_style

        # -- If we still do not have an extracted style then we need
        # -- to report a warning
        if not extracted_style:
            constants.log.warning(
                'Could not extract or locate the style : %s' % given_style
            )
            continue

        # -- Add this extracted data to the compounded style
        compounded_style += '\n' + extracted_style

    # -- We need to combine the kwargs with the defaults
    styling_parameters = constants.STYLE_DEFAULTS.copy()
    styling_parameters.update(kwargs)

    # -- Now that we have compounded all our style information we can cycle
    # -- over it and carry out any replacements
    for regex, replacement in styling_parameters.items():
        regex = re.compile(regex)
        compounded_style = regex.sub(replacement, compounded_style)

    # -- Apply the composed stylesheet
    apply_to.setStyleSheet(compounded_style)


# ------------------------------------------------------------------------------
def getCompoundedStylesheet(widget):
    """
    This will return the entire stylesheet which is affecting this widget.
    The resulting stylesheet will be the widgets stylesheet, and the
    stylesheet of all its parents.

    The order is such that the widgets style is last, and the root level
    widgets style is first - meaning you can apply the returned stylesheet
    to create the exact same result.

    :param widget: Widget to read the style from
    :type widget: QWidget

    :return: stylesheet data
    :rtype: str
    """
    all_data = [widget.styleSheet()]

    while widget.parentWidget():
        # -- Iterate to the parent
        widget = widget.parentWidget()

        # -- Take the stylesheet, or an empty string if there
        # -- is no stylesheet
        all_data.append('' or widget.styleSheet())

    # -- Now we reverse the list, so the base parent comes first and the leaf
    # -- style comes last
    all_data.reverse()

    # -- Finally, we join it all together into a single string
    all_data = '\n'.join(all_data)

    return all_data


# ------------------------------------------------------------------------------
def _getAvailableStyles():
    """
    This will look at all the stylesheet locations on the environment path
    and return a dictionary of them, where the key is the name (without a qss
    or css suffix) and the value is the absolute path.

    Where there is a name clash, the last will always override the first.

    The order of searching is done by the order of paths defined in the
    QUTE_STYLE_PATH environment variable.

    :return:
    """
    styles = dict()

    for location in constants.QUTE_STYLE_LOCATIONS:

        # -- Skip it if it is not accessible
        if not os.path.exists(location):
            continue

        # -- Cycle over the files in the location
        for style_name in os.listdir(location):

            test_name = style_name.lower()

            # -- Only check for qss or css files
            if test_name.endswith('.css') or test_name.endswith('.qss'):
                styles[style_name[:-4]] = os.path.join(location, style_name)

    return styles



from ..vendor import Qt
from . import qApp


# ------------------------------------------------------------------------------
def message(title='Text Request', label='', parent=None, **kwargs):
    """
    Quick and easy access for getting text input. You do not have to have a
    QApplication instance, as this will look for one.

    :return: str, or None
    """
    # -- Ensure we have a QApplication instance
    q_app = qApp()

    answer = Qt.QtWidgets.QMessageBox.information(
        parent,
        title,
        label,
    )

    return True


# ------------------------------------------------------------------------------
def confirmation(title='Text Request', label='', parent=None, **kwargs):
    """
    Quick and easy access for getting text input. You do not have to have a
    QApplication instance, as this will look for one.

    :return: str, or None
    """
    # -- Ensure we have a QApplication instance
    q_app = qApp()

    answer = Qt.QtWidgets.QMessageBox.warning(
        parent,
        title,
        label,
        Qt.QtWidgets.QMessageBox.Yes,
        Qt.QtWidgets.QMessageBox.No,
    )

    if answer == Qt.QtWidgets.QMessageBox.Yes:
        return True

    return False


# ------------------------------------------------------------------------------
def text(title='Text Request', label='', parent=None, **kwargs):
    """
    Quick and easy access for getting text input. You do not have to have a
    QApplication instance, as this will look for one.

    :return: str, or None
    """
    # -- Ensure we have a QApplication instance
    q_app = qApp()

    # -- Get the text
    name, ok = Qt.QtWidgets.QInputDialog.getText(
        parent,
        title,
        label,
        **kwargs
    )

    if not ok:
        return None

    return name


# ------------------------------------------------------------------------------
def filepath(title='Text Request',
                save=True,
                path='',
                filter_='* (*.*)',
                parent=None,
                **kwargs):
    """
    Quick and easy access for getting text input. You do not have to have a
    QApplication instance, as this will look for one.

    :return: str, or None
    """

    # -- Ensure we have a q application instance
    q_app = qApp()

    # -- Ask for the editor location
    func = Qt.QtWidgets.QFileDialog.getOpenFileName

    if save:
        func = Qt.QtWidgets.QFileDialog.getSaveFileName

    value, _ = func(
        parent,
        title,
        path,
        filter_,
    )

    return value


# ------------------------------------------------------------------------------
def folderpath(title='Folder Request', path='', parent=None):
    """
    Quick function for requesting a folder path

    :param title:
    :param path:
    :return:
    """
    q_app = qApp()

    return Qt.QtWidgets.QFileDialog.getExistingDirectory(
        parent,
        title,
        path,
        Qt.QtWidgets.QFileDialog.ShowDirsOnly,
    )




"""
This holds a set of small helper functions
"""
from ..vendor import Qt
from . import layouts


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences,PyPep8Naming
def addLabel(widget, label):
    """
    Adds a label to the widget, returning a layout with the label
    on the left and the widget on the right.

    :param widget: Widget to be given a label
    :type widget: QWidget

    :param label: Label text to show
    :type label: str

    :return: SlimHBoxLayout
    """
    label = Qt.QtWidgets.QLabel(label)
    layout = layouts.slimify(Qt.QtWidgets.QHBoxLayout())

    layout.addWidget(label)
    layout.addSpacerItem(
        Qt.QtWidgets.QSpacerItem(
            10,
            0,
            Qt.QtWidgets.QSizePolicy.Expanding,
            Qt.QtWidgets.QSizePolicy.Fixed,
        ),
    )
    layout.addWidget(widget)

    layout.setStretch(
        2,
        1,
    )

    return layout


# ------------------------------------------------------------------------------
# noinspection PyPep8Naming
def getComboIndex(combo_box, label, ignore_casing=False):
    """
    This will return the index of the first matching label within a combo
    box qwidget.

    If no label is found 0 is returned

    :param combo_box: Widget to iterate through
    :type combo_box: QComboBox

    :param label: The combo label to match against
    :type label: str

    :param ignore_casing: If true, all text matching will be done with no
        consideration of capitalisation. The default is False.
    :type ignore_casing: bool

    :return: int
    """
    # -- Convert the label to lower case if we're ignoring casing
    label = label if not ignore_casing else label.lower()

    # -- Cycling our combo box and test the string
    for i in range(combo_box.count()):

        # -- Get the current item text, and lower the casing if we're
        # -- ignoring the casing. This means we're testing both sides
        # -- of the argument in lower case
        combo_text = combo_box.itemText(i)
        combo_text = combo_text if not ignore_casing else combo_text.lower()

        if combo_text == label:
            return i

    return 0


# ------------------------------------------------------------------------------
def setComboByText(combo_box, label, ignore_casing=False):
    """
    This will return the index of the first matching label within a combo
    box qwidget.

    If no label is found 0 is returned

    :param combo_box: Widget to iterate through
    :type combo_box: QComboBox

    :param label: The combo label to match against
    :type label: str

    :param ignore_casing: If true, all text matching will be done with no
        consideration of capitalisation. The default is False.
    :type ignore_casing: bool

    :return: int
    """
    idx = getComboIndex(combo_box, label, ignore_casing=ignore_casing)
    combo_box.setCurrentIndex(idx)




from ..vendor import Qt


# ------------------------------------------------------------------------------
class HorizontalDivider(Qt.QtWidgets.QFrame):

    # ------------------------------------------------------------------------------
    def __init__(self, height=2):
        super(HorizontalDivider, self).__init__()
        self.setFrameShape(Qt.QtWidgets.QFrame.HLine)
        self.setFrameShadow(Qt.QtWidgets.QFrame.Sunken)
        self.setStyleSheet("background-color: rgb(20,20,20)")
        self.setFixedHeight(height)



"""
This module is specifically intended for use when in environments where
you're actively trying to share/develop tools across multiple applications
which support PyQt, PySide or PySide2. 

The premise is that you can request the main application window using 
a common function regardless of the actual application - making it trivial
to implement a tool which works in multiple host applications without any
bespoke code.

The current list of supported applications are:

    * Native Python
    * Maya
    * 3dsmax
    * Motion Builder

"""
import sys

from ..vendor import Qt


# ------------------------------------------------------------------------------
def get_host():

    global HOST

    if HOST:
        return HOST

    if 'maya.exe' in sys.executable or 'mayapy.exe' in sys.executable:
        HOST = 'Maya'
        return HOST

    if 'motionbuilder.exe' in sys.executable:
        HOST = 'Mobu'
        return HOST

    if '3dsmax.exe' in sys.executable:
        HOST = 'Max'
        return HOST

    houdini_execs = [
        'houdini.exe',
        'houdinifx.exe',
        'houdinicore.exe',
    ]
    if any(houdini_exec in sys.executable for houdini_exec in houdini_execs):
        HOST = 'Houdini'
        return HOST

    return 'Pure'


# ------------------------------------------------------------------------------
# noinspection PyPep8Naming
def mainWindow():
    """
    Returns the main window regardless of what the host is
    
    :return: 
    """
    return HOST_MAPPING[get_host()]()


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences,PyPep8Naming
def returnNativeWindow():
    for candidate in Qt.QtWidgets.QApplication.topLevelWidgets():
        if isinstance(candidate, Qt.QtWidgets.QMainWindow):
            return candidate


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences,PyPep8Naming
def _findWindowByTitle(title):
    # -- Find the main application window
    for candidate in Qt.QtWidgets.QApplication.topLevelWidgets():
        # noinspection PyBroadException
        try:
            if title in candidate.windowTitle():
                return candidate
        except: pass


# ------------------------------------------------------------------------------
# noinspection PyPep8Naming
def returnModoMainWindow():
    pass


# ------------------------------------------------------------------------------
# noinspection PyPep8Naming
def returnMaxMainWindow():
    return _findWindowByTitle('Autodesk 3ds Max')


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences,PyPep8Naming
def returnMayaMainWindow():
    from maya import OpenMayaUI as omui
    from shiboken2 import wrapInstance
    
    try:
        return wrapInstance(long(omui.MQtUtil.mainWindow()), Qt.QtWidgets.QWidget)

    except:
        pass

    return wrapInstance(int(omui.MQtUtil.mainWindow()), Qt.QtWidgets.QWidget)


# ------------------------------------------------------------------------------
# noinspection PyPep8Naming
def returnHoudiniMainWindow():
    import hou
    return hou.qt.mainWindow()


# ------------------------------------------------------------------------------
# noinspection PyPep8Naming
def returnMobuMainWindow():
    return _findWindowByTitle('MotionBuilder 20')


# ------------------------------------------------------------------------------
HOST = None
HOST_MAPPING = dict(
    Maya=returnMayaMainWindow,
    Max=returnMaxMainWindow,
    Modo=returnModoMainWindow,
    Mobu=returnMobuMainWindow,
    Pure=returnNativeWindow,
    Houdini=returnHoudiniMainWindow,
)



from ._core import *

from . import derive
from . import designer
from . import events
from . import layouts
from . import menus
from . import pixmaps
from . import styling
from . import widgets
from . import windows
from . import request
from . import sizing



from ..vendor.Qt.QtGui import QPixmap, QImage
from ..vendor.Qt.QtCore import QByteArray, QDataStream, QIODevice

from ..vendor import Qt


# --------------------------------------------------------------------------
def toGrayscale(pixmap):
    """
    Creates a new pixmap which is a grayscale version of the given
    pixmap

    :param pixmap: QPixmap

    :return: QPixmap
    """
    # -- Get an image object
    image = pixmap.toImage()

    # -- Cycle the pixels and convert them to grayscale
    for x in range(image.width()):
        for y in range(image.height()):

            # -- Grayscale the pixel
            gray = Qt.QtGui.qGray(
                image.pixel(
                    x,
                    y,
                ),
            )

            # -- Set the pixel back into the image
            image.setPixel(
                x,
                y,
                Qt.QtGui.QColor(
                    gray,
                    gray,
                    gray,
                ).rgb()
            )

    # -- Re-apply the alpha channel
    image.setAlphaChannel(
        pixmap.toImage().alphaChannel()
    )

    # -- Return the pixmap
    return Qt.QtGui.QPixmap.fromImage(image)



from ..vendor import Qt
from ..vendor import scribble


# ------------------------------------------------------------------------------
# noinspection PyPep8Naming,PyUnresolvedReferences
class MemorableWindow(Qt.QtWidgets.QMainWindow):

    # -- How many pixels of a window we expect to see
    # -- before we re-position it
    _SCREEN_BUFFER = 12

    # --------------------------------------------------------------------------
    def __init__(self, identifier=None, offsetX=7, offsetY=32, *args, **kwargs):
        super(MemorableWindow, self).__init__(*args, **kwargs)

        self._offsetX = offsetX
        self._offsetY = offsetY

        # -- If we're given an id, set this object name
        if identifier:
            self.setObjectName(identifier)

            # -- If there is any scribble data for this id we should
            # -- update this windows geometry accordingly.
            try:
                self.restoreSize()
            except:
                pass

            try:
                self.restore()
            except:
                pass

    def restore(self):
        settings = Qt.QtCore.QSettings('quteSettings', self.objectName())
        self.restoreState(settings.value('windowState'))

    def save(self):
        settings = Qt.QtCore.QSettings('quteSettings', self.objectName())
        settings.setValue('windowState', self.saveState())
        self.storeSize()

    def closeEvent(self, event):
        self.save()
        super(MemorableWindow, self).closeEvent(event)

    # --------------------------------------------------------------------------
    def restoreSize(self):
        stored_data = scribble.get(self.objectName())

        # -- Get the screen rect, as we need to make sure
        # -- we're not about to open the window offscreen
        screen_rect = Qt.QtWidgets.QApplication.desktop().availableGeometry()

        if 'geometry' in stored_data:
            geom = stored_data.get(
                'geometry',
                [
                    300,
                    300,
                    400,
                    400,
                ],
            )

            # -- Ensure the window is not too far off to the left
            # -- side of the screen, and ensure that we see at least
            # -- some of the window
            if geom[0] + (geom[2] - self._SCREEN_BUFFER) < 0:
                geom[0] = 300

            # -- Ensure that its not too far to the right, making
            # -- sure that we see at least some of the window
            if geom[0] + self._SCREEN_BUFFER > screen_rect.width():
                geom[0] = 300

            # -- Check if the window is above the available monitor space
            # -- we dont need to use the buffer on this, because we dont
            # -- want the title bar to go too high
            if geom[1] < 0:
                geom[1] = 300

            # -- Check if the window is below the monitor with at least
            # -- a small amount of the title bar to grab
            if geom[1] + self._SCREEN_BUFFER > screen_rect.height():
                geom[1] = 300

            self.setGeometry(
                geom[0],
                geom[1],
                geom[2],
                geom[3],
            )

    # --------------------------------------------------------------------------
    def storeSize(self):
        stored_data = scribble.get(self.objectName())
        stored_data['geometry'] = [
            self.pos().x() + self._offsetX,
            self.pos().y() + self._offsetY,
            self.width(),
            self.height(),
        ]
        stored_data.save()

    # --------------------------------------------------------------------------
    def resizeEvent(self, event):
        self.storeSize()
        super(MemorableWindow, self).resizeEvent(event)

    # --------------------------------------------------------------------------
    def moveEvent(self, event):
        self.storeSize()
        super(MemorableWindow, self).moveEvent(event)

    # --------------------------------------------------------------------------
    def hideEvent(self, event):
        self.storeSize()
        super(MemorableWindow, self).hideEvent(event)
